"""The class for testing models and reporting results

A collection of methods to simplify your code.

# license MIT
# author Alessandra Bilardi <alessandra.bilardi@gmail.com>
# see https://github.com/bilardi/smltk for details
"""
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_recall_fscore_support
#%pip install mlxtend --upgrade
from mlxtend.evaluate import bias_variance_decomp
import seaborn as sns
import numpy as np
import pickle

class Metrics():
    def split_tuples(self, tuples):
        """
        splits tuples of sample and its target in the relative lists
            Arguments:
                tuples (list[tuple]): list of tuples with sample and its target
            Returns:
                tuple of list of samples and list of targets
        """
        X_test = []
        y_test = []
        for sample, target in tuples:
            X_test.append(sample)
            y_test.append(target)
        return X_test, y_test

    def prediction(self, model, method, X_test, y_test = []):
        """
        predicts with your model
            Arguments:
                model (obj): object of your model
                X_test (list[]|list[tuple]): list of samples or list of tuples with sample and its target
                y_test (list[]): list of targets
            Returns:
                tuple of list of targets and list of predictions
        """
        y_pred = []
        if type(X_test[0]) is tuple:
            X_test, y_test = self.split_tuples(X_test)
        predictor = getattr(model, method)
        for sample in X_test:
            y_pred.append(predictor(sample))
        return y_test, y_pred

    def create_confusion_matrix(self, y_test, y_pred):
        """
        creates and prints confusion matrix
            Arguments:
                y_test (list[]): list of targets
                y_pred (list[]): list of predictions
            Returns:
                confusion matrix
        """
        matrix = confusion_matrix(y_test, y_pred)
        tick_labels = np.unique(np.array(y_test + y_pred))
        sns.heatmap(matrix.T, square=True, annot=True, fmt='d', cbar=False, xticklabels = tick_labels, yticklabels = tick_labels)
        return matrix

    def get_classification_metrics(self, params = []):
        """
        gets classification metrics
            Arguments:
                params (dict):
                    model (obj): object of your model
                    X_train (list[]|list[tuple]): list of samples or list of tuples with sample and its target
                    y_train (list[]): list of targets
                    X_test (list[]|list[tuple]): list of samples or list of tuples with sample and its target
                    y_test (list[]): list of targets
                    y_pred (list[]): list of predictions
                    loss (str): parameter of bias_variance_decomp, default mse
                    num_rounds (int): parameter of bias_variance_decomp, default 200
                    random_seed (int): parameter of bias_variance_decomp, default 3
            Returns:
                dictionary with MSE, Bias, Variance, Accuracy, Precision, Recall, Fscore
        """
        if 'loss' not in params:
            params['loss'] = 'mse'
        if 'num_rounds' not in params:
            params['num_rounds'] = 200
        if 'random_seed' not in params:
            params['random_seed'] = 3
        fit_exists = hasattr(params['model'], 'fit') and callable(getattr(params['model'], 'fit'))
        predict_exists = hasattr(params['model'], 'predict') and callable(getattr(params['model'], 'predict'))
        if fit_exists == True or predict_exists == True:
            mse, bias, variance = bias_variance_decomp(params['model'], params['X_train'], params['y_train'], params['X_test'], params['y_test'], loss=params['loss'], num_rounds=params['num_rounds'], random_seed=params['random_seed'])
        else:
            mse, bias, variance = (0, 0, 0)
        accuracy = accuracy_score(params['y_test'], params['y_pred'])
        report = precision_recall_fscore_support(params['y_test'], params['y_pred'])
        return {'MSE': mse, 'Bias': bias, 'Variance': variance, 'Accuracy': accuracy, 'Precision': report[0], 'Recall': report[1], 'Fscore': report[2]}

    def print_metrics(self, metrics):
        """
        prints metrics
            Arguments:
                metrics (dict): dictionary of metrics with their value
            Returns:
                only the print of metrics
        """
        for metric in metrics.keys():
            print(metric, metrics[metric])

    def save_model(self, model, filename):
        """
        saves model
            Arguments:
                model (obj): object of your model
                filename (str): pathname and filename where you want to save your model
        """
        with open(filename, 'wb') as pickle_file:
            pickle.dump(model, pickle_file)

    def resume_model(self, filename):
        """
        resumes model
            Arguments:
                filename (str): pathname and filename where you want to save your model
            Returns:
                object of your model
        """
        with open(filename, 'rb') as pickle_file:
          return pickle.load(pickle_file)



"""The classes for data preparation with nltk, sklearn and more

    A collection of methods to simplify your code.

"""
import re
import string
import nltk
nltk.download('wordnet')
nltk.download('averaged_perceptron_tagger')
nltk.download('stopwords')
from nltk.corpus import stopwords
from nltk.corpus import wordnet as wn
from nltk.stem import WordNetLemmatizer
from collections import defaultdict
from collections import Counter
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer

class Ntk():
    """
    The class Ntk contains the Natural Language Processing tool kit.
    It needs the property named **params** (dict)
        :param language (str): default is english
        :param lemmatizer (obj): obj with method lemmatize() like WordNetLemmatizer()
        :param min_length (int): default is 1, so all words will be used
        :param stop_words (list[str]): default is stopwords.words()
        :param tag_map (dict): default contains J, V, R

    Here's an example:

        >>> from smltk.preprocessing import Ntk
        >>> doc = 'Good case, Excellent value.'
        >>> ntk = Ntk()
        >>> get_doc_cleaned = ntk.get_doc_cleaned(doc)
        >>> print(get_doc_cleaned)
        good case excellent value
    """
    language = 'english'
    lemmatizer = None
    min_length = 1
    stop_words = []
    tag_map = {}

    def __init__(self, params = []):
        if 'language' in params:
            self.language = params['language']
        if 'lemmatizer' in params:
            self.lemmatizer = params['lemmatizer']
        else:
            self.lemmatizer = WordNetLemmatizer()
        if 'min_length' in params:
            self.min_length = params['min_length']
        if 'stop_words' in params:
            self.stop_words = set(params['stop_words'])
        else:
            self.stop_words = set(stopwords.words(self.language))
        if 'tag_map' in params:
            self.tag_map = params['tag_map']
        else:
            self.tag_map = defaultdict(lambda : wn.NOUN)
            self.tag_map['J'] = wn.ADJ
            self.tag_map['V'] = wn.VERB
            self.tag_map['R'] = wn.ADV
        if not len(self.tag_map):
            for pos in ['E', 'D', 'J', 'N', 'V', 'R']:
                self.tag_map[pos] = wn.NOUN

    def word_tokenize(self, doc):
        """
        splits document in each word
            Arguments:
                doc (str): text
            Returns:
                list of words
        """
        return nltk.word_tokenize(doc)

    def tokenize_and_clean_doc(self, doc):
        """
        tokenizes doc
        and filters tokens from punctuation, numbers, stop words, and words <= min_length
        and changes upper characters in lower characters
            Arguments:
                doc (str): text
            Returns:
                list of words filtered
        """
        # split into tokens by white space
        tokens = self.word_tokenize(doc)
        # prepare regex for char filtering
        punctuation = re.compile('[%s]' % re.escape(string.punctuation))
        # remove punctuation from each word
        tokens = [punctuation.sub('', token) for token in tokens]
        # remove remaining tokens with no alphabetic characters
        tokens = [token for token in tokens if token.isalpha()]
        # convert in lower case
        tokens = [token.lower() for token in tokens]
        # filter out stop words
        tokens = [token for token in tokens if not token in self.stop_words]
        # filter out short tokens
        tokens = [token for token in tokens if len(token) >= self.min_length]
        return tokens

    def lemmatize(self, tokens):
        """
        lemmatizes tokens
            Arguments:
                tokens (list[str]): list of words
            Returns:
                list of tokens lemmatized
        """
        # lemmatization
        lemmas = [self.lemmatizer.lemmatize(token, self.tag_map[tag[0]]) for token, tag in nltk.pos_tag(tokens)]
        return lemmas

    def get_tokens_cleaned(self, doc, is_lemma = True):
        """
        tokenizes doc
        and filters tokens from punctuation, numbers, stop words, and words <= min_length
        and changes upper characters in lower characters
        and if is_lemma == True, also it lemmatizes
            Arguments:
                doc (str): text
                is_lemma (bool): default is True
            Returns:
                list of tokens cleaned
        """
        tokens = self.tokenize_and_clean_doc(doc)
        if is_lemma == True:
            tokens = self.lemmatize(tokens)
        return tokens

    def __call__(self, doc):
        """
        alias of get_tokens_cleaned for using in models
            Arguments:
                doc (str): text
            Returns:
                list of tokens cleaned
        """
        return self.get_tokens_cleaned(doc)

    def get_doc_cleaned(self, doc, is_lemma = True):
        """
        filters doc from punctuation, numbers, stop words, and words <= min_length
        and changes upper characters in lower characters
        and if is_lemma == True, also it lemmatizes
            Arguments:
                doc (str): text
                is_lemma (bool): default is True
            Returns:
                string cleaned
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        return ' '.join(map(str, tokens))

    def add_doc_to_vocab(self, doc, vocab, is_lemma = True):
        """
        adds tokens of that doc to vocabulary and updates vocabulary
            Arguments:
                doc (str): text
                vocab (list[str]): list of tokens
                is_lemma (bool): default is True
            Returns:
                list of tokens of that doc and vocab updated
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        vocab.update(tokens)
        return tokens

    def create_vocab_from_docs(self, docs, is_lemma = True):
        """
        creates vocabulary from list of docs
            Arguments:
                docs (list[str]): list of texts
                is_lemma (bool): default is True
            Returns:
                list of tokens
        """
        vocab = Counter()
        [self.add_doc_to_vocab(doc, vocab, is_lemma) for doc in docs]
        return vocab

    def get_stats_vocab(self, vocab, min_occurance = 1):
        """
        gets statistics of vocabulary
            Arguments:
                vocab (list[str]): list of tokens
                min_occurance (int): minimum occurance considered
            Returns:
                tuple of tokens number with >= min_occurance and total tokens number
        """
        tokens = [ key for key, count in vocab.items() if count >= min_occurance]
        partial = len(tokens)
        total = len(vocab)
        return partial, total

    def create_tuples(self, docs = [], target = []):
        """
        creates tuples with sample and its target
            Arguments:
                docs (list[str]): list of texts
                target (list[str]): list of targets
            Returns:
                list of tuples with sample and its target
        """
        return list(zip(docs, target))

    def create_vocab_from_tuples(self, tuples, is_lemma = True):
        """
        creates vocabulary from list of tuples
            Arguments:
                tuples (list[tuples]): list of tuples with sample and its target
                is_lemma (bool): default is True
            Returns:
                list of tokens
        """
        vocab = Counter()
        [self.add_doc_to_vocab(doc, vocab, is_lemma) for doc, target in tuples]
        return vocab

    def find_features(self, doc, is_lemma = True):
        """
        finds features
            Arguments:
                doc (str): text
                is_lemma (bool): default is True
            Returns:
                dictionary of tokens cleaned
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        features = {}
        for token in tokens:
            features[token] = True
        return features

    def create_features(self, tuples, is_lemma = True):
        """
        creates features
            Arguments:
                tuples (list[tuples]): list of tuples with sample and its target
                is_lemma (bool): default is True
            Returns:
                list of tuples with features and relative target
        """
        return [(self.find_features(doc, is_lemma), target) for (doc, target) in tuples]

    def vectorize_docs(self, docs, is_count = True, is_lemma = True):
        """
        vectorizes docs
            Arguments:
                docs (list[str]): list of texts
                is_count (bool): default is True
                is_lemma (bool): default is True
            Returns:
                list of scipy.sparse.csr.csr_matrix, one for each doc
        """
        if is_count == True:
            vectorizer = CountVectorizer()
        else:
            vectorizer = TfidfVectorizer()
        return vectorizer.fit_transform(docs)



"""Simple Machine Learning Tool Kit package

This package contains the modules to simplify your code for your machine learning processes.

It is part of the educational repositories (https://github.com/pandle/materials)
to learn how to write stardard code and common uses of the TDD.

Package contents two classes to manage data preprocessing, testing and reporting.

    >>> import smltk
    >>> help(smltk)
    >>> import smltk.metrics as Metrics
    >>> help(Metrics)

# license MIT
# support https://github.com/bilardi/smltk/issues
"""
__version__ = '0.0.3'
__author__ = 'Alessandra Bilardi'



import unittest
from smltk.preprocessing import Ntk
import nltk
nltk.download('punkt')
from nltk.stem import WordNetLemmatizer
import scipy

class TestNlp(unittest.TestCase, Ntk):
    ntk = None
    doc = 'Good case, Excellent value. I am agree. There is a mistake. Item Does Not Match Picture.'
    docs = []
    target = []
    tuples = []
    default_doc_filtered = 'good case excellent value agree mistake item match picture'
    without_stop_words_doc_filtered = 'good case excellent value i be agree there be a mistake item do not match picture'
    without_stop_words_with_min_length_doc_filtered = 'good case excellent value be agree there be mistake item do not match picture'
    without_stop_words_without_tag_map_doc_filtered = 'good case excellent value i am agree there is a mistake item doe not match picture'
    def __init__(self, *args, **kwargs):
        self.ntk = Ntk()
        self.docs = nltk.sent_tokenize(self.doc)
        self.target = [1, 1, 0, 0]
        self.tuples = self.ntk.create_tuples(self.docs, self.target)
        unittest.TestCase.__init__(self, *args, **kwargs)

    def test_doc_cleaned(self):
        self.assertEqual(self.ntk.get_doc_cleaned(self.doc), self.default_doc_filtered)

        vocab_token = self.ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 9)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token, 2), (0, 9))

        vocab_lemma = self.ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 9)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_lemma), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_lemma, 2), (0, 9))

        features_token = self.ntk.create_features(self.tuples, False)
        self.assertEqual(features_token[1], ({'agree': True}, 1))

        features_lemma = self.ntk.create_features(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'agree': True}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(self.ntk.get_doc_cleaned(doc, False))
        tuples = self.ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = self.ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token, 2), (0, 9))

    def test_doc_cleaned_with_lemmatizer(self):
        ntk = Ntk({'lemmatizer': WordNetLemmatizer()})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.default_doc_filtered)

    def test_doc_cleaned_without_stop_words(self):
        ntk = Ntk({'stop_words': []})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.without_stop_words_doc_filtered)

        vocab_token = ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 16)
        self.assertEqual(vocab_token['am'], 1)
        self.assertEqual(vocab_token['is'], 1)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        vocab_lemma = ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 15)
        self.assertEqual(vocab_lemma['be'], 2)
        self.assertEqual(vocab_lemma['do'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma), (15, 15))
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma, 2), (1, 15))

        features_token = ntk.create_features(self.tuples, False)
        self.assertEqual(features_token[1], ({'agree': True, 'am': True, 'i': True}, 1))

        features_lemma = ntk.create_features(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'agree': True, 'be': True, 'i': True}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, False))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, True))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (15, 15))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (1, 15))

    def test_doc_cleaned_without_stop_words_with_min_length(self):
        ntk = Ntk({'stop_words': [], 'min_length': 2})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.without_stop_words_with_min_length_doc_filtered)

        vocab_token = ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 14)
        self.assertEqual(vocab_token['am'], 1)
        self.assertEqual(vocab_token['is'], 1)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (14, 14))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 14))

        vocab_lemma = ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 13)
        self.assertEqual(vocab_lemma['be'], 2)
        self.assertEqual(vocab_lemma['do'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma), (13, 13))
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma, 2), (1, 13))

        features_token = ntk.create_features(self.tuples, False)
        self.assertEqual(features_token[1], ({'agree': True, 'am': True}, 1))

        features_lemma = ntk.create_features(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'agree': True, 'be': True}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, False))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (14, 14))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 14))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, True))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (13, 13))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (1, 13))

    def test_doc_cleaned_without_stop_words_without_tag_map(self):
        ntk = Ntk({'stop_words': [], 'tag_map': {}})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.without_stop_words_without_tag_map_doc_filtered)

        vocab_token = ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 16)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        vocab_lemma = ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 16)
        self.assertEqual(vocab_lemma['doe'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma, 2), (0, 16))

        features_token = ntk.create_features(self.tuples, False)
        self.assertEqual(features_token[1], ({'agree': True, 'am': True, 'i': True}, 1))

        features_lemma = ntk.create_features(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'agree': True, 'am': True, 'i': True}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, False))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, True))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

    def test_vectorize_docs(self):
        X_test = self.ntk.vectorize_docs(self.docs)
        self.assertEqual(type(X_test[0]), scipy.sparse.csr.csr_matrix)
        X_test = self.ntk.vectorize_docs(self.docs, False)
        self.assertEqual(type(X_test[0]), scipy.sparse.csr.csr_matrix)

if __name__ == '__main__':
    unittest.main()


import setuptools
import smltk

setuptools.setup(
    name="smltk",
    version=smltk.__version__,
    author=smltk.__author__,
    author_email="alessandra.bilardi@gmail.com",
    description="Simple Machine Learning Tool Kit package",
    long_description=open('README.rst').read(),
    long_description_content_type="text/x-rst",
    url="https://smltk.readthedocs.io/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    project_urls={
        "Source":"https://github.com/bilardi/smltk",
        "Bug Reports":"https://github.com/bilardi/smltk/issues",
        "Funding":"https://donate.pypi.org",
    },
)
